@@ -467,6 +467,8 @@ let default_prompt =
 
 let prompt = ref default_prompt
 
+let edit_mode= ref LTerm_editor.Default
+
 let () =
   Hashtbl.add Toploop.directive_table "utop_prompt_simple"
     (Toploop.Directive_none
@@ -488,7 +490,17 @@ let () =
     (Toploop.Directive_none
        (fun () ->
          set_profile Dark;
-         prompt := default_prompt))
+         prompt := default_prompt));
+
+  Hashtbl.add Toploop.directive_table "edit_mode_default"
+    (Toploop.Directive_none
+       (fun () ->
+         edit_mode:= LTerm_editor.Default));
+
+  Hashtbl.add Toploop.directive_table "edit_mode_vi"
+    (Toploop.Directive_none
+       (fun () ->
+         edit_mode:= LTerm_editor.Vi))
 
 (* +-----------------------------------------------------------------+
    | Help                                                            |
@@ -743,6 +755,41 @@ let () =
     (Toploop.Directive_string
        (fun str -> require (split_words str)))
 
+(* +-----------------------------------------------------------------+
+   | Backports                                                       |
+   +-----------------------------------------------------------------+ *)
+
+
+let try_finally ~always work=
+#if OCAML_VERSION >= (4, 08, 0)
+    Misc.try_finally ~always work
+#else
+    Misc.try_finally work always
+#endif
+
+let use_output command =
+  let fn = Filename.temp_file "ocaml" "_toploop.ml" in
+  try_finally ~always:(fun () ->
+    try Sys.remove fn with Sys_error _ -> ())
+    (fun () ->
+       match
+         Printf.ksprintf Sys.command "%s > %s"
+           command
+           (Filename.quote fn)
+       with
+       | 0 ->
+         ignore (Toploop.use_file Format.std_formatter fn : bool)
+       | n ->
+         Format.printf "Command exited with code %d.@." n)
+
+let () =
+  let name = "use_output" in
+  if not (Hashtbl.mem Toploop.directive_table name) then
+    Hashtbl.add
+      Toploop.directive_table
+      name
+      (Toploop.Directive_string use_output)
+
 (* +-----------------------------------------------------------------+
    | Initialization                                                  |
    +-----------------------------------------------------------------+ *)
