@@ -80,36 +80,38 @@ let load () =
       styles.style_foreground <- LTerm_resources.get_color "foreground" res;
       styles.style_background <- LTerm_resources.get_color "background" res;
       styles.style_cursor <- LTerm_resources.get_color "cursor" res;
-      (match String.lowercase (LTerm_resources.get "profile" res) with
+      (match String.lowercase_ascii (LTerm_resources.get "profile" res) with
          | "light" -> UTop.set_profile UTop.Light
          | "dark" -> UTop.set_profile UTop.Dark
          | "" -> ()
          | str -> raise (LTerm_resources.Error (Printf.sprintf "invalid profile %S" str)));
       UTop_private.error_style := styles.style_error;
       UTop_private.autoload := LTerm_resources.get_bool "autoload" res <> Some false;
+      (match LTerm_resources.get "external-editor" res with
+       | "" -> ()
+       | s  -> UTop.set_external_editor s);
       return ())
     (function
     | Unix.Unix_error(Unix.ENOENT, _, _) ->
         return ()
-    | Unix.Unix_error (error, func, arg) ->
+    | Unix.Unix_error (error, func, _arg) ->
         Lwt_log.error_f "cannot load styles from %S: %s: %s" fn func (Unix.error_message error)
     | exn -> Lwt.fail exn)
 
-let rec stylise_filter_layout stylise tokens =
-  match tokens with
-    | [] ->
-        []
-    | (Comment (Comment_reg, _), loc) :: tokens ->
+let stylise_filter_layout stylise tokens =
+  let aux acc = function
+    | (Comment (Comment_reg, _), loc) ->
         stylise loc styles.style_comment;
-        stylise_filter_layout stylise tokens
-    | (Comment (Comment_doc, _), loc) :: tokens ->
+        acc
+    | (Comment (Comment_doc, _), loc) ->
         stylise loc styles.style_doc;
-        stylise_filter_layout stylise tokens
-    | (Blanks, loc) :: tokens ->
+        acc
+    | (Blanks, loc) ->
         stylise loc styles.style_blanks;
-        stylise_filter_layout stylise tokens
-    | x :: tokens ->
-        x :: stylise_filter_layout stylise tokens
+        acc
+    | x -> x :: acc
+  in
+  List.rev (List.fold_left aux [] tokens)
 
 let rec stylise_rec stylise tokens =
   match tokens with
@@ -128,11 +130,11 @@ let rec stylise_rec stylise tokens =
     | (Uident id, loc) :: tokens when String_set.mem id !UTop.keywords ->
         stylise loc styles.style_keyword;
         stylise_rec stylise tokens
-    | (Uident id, loc1) :: (Symbol ".", loc2) :: tokens ->
+    | (Uident _id, loc1) :: (Symbol ".", loc2) :: tokens ->
         stylise loc1 styles.style_module;
         stylise loc2 styles.style_symbol;
         stylise_rec stylise tokens
-    | (Uident id, loc) :: tokens ->
+    | (Uident _id, loc) :: tokens ->
         stylise loc styles.style_ident;
         stylise_rec stylise tokens
     | (Constant _, loc) :: tokens ->
